#include <config.h>

module hmc_gluon_class
!
!= Definition of HMC gluon action
!
! This action contains RG action (1x1-Plaquette + 1x2-Rectangular).
!
!== $Id: hmc_gluon_class.F90,v 1.24 2011/09/25 06:12:47 ishikawa Exp $
!
  use lattice_class
  use error_class
  use logfile_class
  use hmc_logfile_class
  use action_base_class
  use field_gauge_class
  implicit none
  private
  public :: vfield_gluon_wog  ! public for gauge mementum
  public :: vfield_gluon_wg   ! public for gauge link
  public :: tfield_gluon_wg   ! public for 2-gauge link
  public :: plaquette
  public :: reunitarize
  public :: get_force_norm
  public :: set_identity
  public :: make_two_links
  public :: copy_sfg_time
  public :: print_copy_sfg_statistics
  public :: clear

  public :: gluon_rg, gluon_rg_sf
  public :: hmc_gluon
  public :: get_id
  public :: set_id
  public :: wloops_obj
  public :: new
  public :: read
  public :: save_config
  public :: print
  public :: delete
  public :: delta_p
  public :: delta_u
  public :: force
  public :: hamil
  public :: update_u
  public :: update_p

  type(hmc_logfile),  save :: m_reunit_log
  character(CHARLEN), save :: m_reunit_log_fname="reunit."

  type wloops_obj
  !
  ! contains Wilson loop/plaquette values
  !
    real(DP) :: PLQ=0.0_DP
    ! PLQ = Sum[Re[1x1],vol,direction]/direction/Vol
    real(DP) :: RCT=0.0_DP
    !
    ! RCT = Sum[Re[2x1+1x2],vol,direction]/direction/Vol
    !     = 2 Sum[Re[1x2],vol,direction]/direction/Vol
    !
  end type

  type, extends(action_parameters) :: gluon_rg
    real(DP)                     :: beta          ! beta
    real(DP)                     ::  cRG(0:3)     ! RG-gauge improvement coefficients
    real(DP)                     :: cRG0(0:3)     ! RG-gauge improvement coefficients (original input)
    real(DP)                     :: u0            !  tadpole improvement coefficients
  end type

  type, extends(gluon_rg) :: gluon_rg_sf
    real(DP) :: ct
    real(DP) :: ct_rg
  end type

  type :: hmc_gluon
  !
  ! Gluon action parameters
  !
    character(CHARLEN),  public  :: file_name     ! filename for action parameters
#ifdef _SF
    type(gluon_rg_sf),   public  :: gluon
#else
    type(gluon_rg),      public  :: gluon
#endif
    type(wloops_obj),    public  :: WLP
    integer,             private :: myid  = 0
    integer,             private :: depth = 0
  end type

  interface new
  !
  ! initialize gluon action parameters
  !
    module procedure new_gluon
  end interface

  interface delete
  !
  ! delete gluon action parameters
  !
    module procedure delete_gluon
  end interface

  interface read
  !
  ! read gauge action parameters
  !
    module procedure read_gluon
  end interface

  interface save_config
  !
  ! save gluon action parameters in configuration file for measurement
  !
    module procedure save_config_gluon
  end interface

  interface print
  !
  ! print gluon action parameters
  !
    module procedure print_gluon
  end interface

  interface force
  !
  ! HMC force for gluon action
  !
    module procedure force_gluon
  end interface

  interface hamil
  !
  ! HMC Hamiltonian for gluon action
  !
    module procedure hamil_gluon
  end interface

  interface reunitarize
    module procedure reunit_u_gluon
  end interface

contains

subroutine new_gluon(this,id)
!
! Initialize/create gluon action
!
  implicit none
  type(hmc_gluon), intent(inout) :: this
  integer,         intent(in)    :: id
  this%file_name = REPEAT(" ",LEN(this%file_name))
  this%gluon%beta      = 0.0_DP
  this%gluon%cRG(:)    = 0.0_DP
  this%gluon%cRG0(:)   = 0.0_DP
  this%gluon%u0        = 1.0_DP
#ifdef _SF
  this%gluon%ct        = 0.0_DP
  this%gluon%ct_rg     = 0.0_DP
#endif
  this%myid = id
  call new(m_reunit_log,TRIM(m_reunit_log_fname))
  return
end subroutine

subroutine delete_gluon(this)
!
! delete gluon action
!
  implicit none
  type(hmc_gluon), intent(inout) :: this
  call delete(m_reunit_log)
  return
end subroutine

subroutine print_gluon(this)
!
! print out gluon action parameters on display
!
  implicit none
  type(hmc_gluon), intent(in) :: this

  if (nodeid==0) then
    write(*,'(16X,"           Beta :",F12.6)') this%gluon%beta
  
#ifdef _RGGAUGE
    write(*,'(16X," RG-gauge  cRG0 :",E24.16)')this%gluon%cRG(0)
    write(*,'(16X,"           cRG1 :",E24.16)')this%gluon%cRG0(1)
    write(*,'(16X,"             u0 :",E24.16)')this%gluon%u0
    write(*,'(16X,"      cRG1/u0^2 :",E24.16)')this%gluon%cRG(1)
#endif

#ifdef _SF
    write(*,'(16X," SF boundary ct :",F12.6)') this%gluon%ct
#ifdef _RGGAUGE
    write(*,'(16X,"          ct_RG :",F12.6)') this%gluon%ct_rg
#endif
#endif

    write(*,'(16X," MD Force depth :",I3)')    this%depth
  endif

  return
end subroutine

subroutine read_gluon(this)
!
! Read gluon action parameters (couplings etc.) form formatted file
!
  use comlib
  use file_tools_class
  use print_status_class
  implicit none
  type(hmc_gluon), intent(inout) :: this
  integer :: iout
  character(CHARLEN) :: char

  if (nodeid == 0) then
    write(*,'(80("-"))')
  endif

  iout = search_free_file_unit()

  if (nodeid==0) then
  open(iout,file=this%file_name,status='old',form='formatted')
  read(iout,*)this%gluon%beta
#ifdef _RGGAUGE
  read(iout,*)this%gluon%cRG0(1)
  read(iout,*)this%gluon%u0
#endif
#ifdef _SF
  read(iout,*)this%gluon%ct
#ifdef _RGGAUGE
  read(iout,*)this%gluon%ct_rg
#endif
#endif
  read(iout,*)this%depth
  close(iout)
  endif

#ifndef _singlePU
  call comlib_bcast(this%gluon%beta,0)
  call comlib_bcast(this%gluon%cRG0(1),0)
  call comlib_bcast(this%gluon%u0,0)
  call comlib_bcast(this%depth,0)
#ifdef _SF
  call comlib_bcast(this%gluon%ct,0)
  call comlib_bcast(this%gluon%ct_rg,0)
#endif
#endif

#ifdef _RGGAUGE
  this%gluon%cRG(0)=1.0d0-8.0d0*this%gluon%cRG0(1)
  this%gluon%cRG(1)=this%gluon%cRG0(1)/this%gluon%u0**2
#endif

  write(char,'(1X,A," for gauge action : read OK ",I3)')  &
                        &      TRIM(this%file_name),nodeid
  call print_status(char)

  return
end subroutine

subroutine save_config_gluon(this,iout)
!
! Save gauge action parameters (couplings etc.) to measurement config file 
! in unformatted form.
!
  implicit none
  type(hmc_gluon), intent(in) :: this
  integer, intent(in) :: iout
  write(iout)this%gluon%beta
#ifdef _RGGAUGE
  write(iout)this%gluon%cRG0(1)
  write(iout)this%gluon%u0
#endif
#ifdef _SF
  write(iout)this%gluon%ct
#ifdef _RGGAUGE
  write(iout)this%gluon%ct_rg
#endif
#endif
  return
end subroutine

subroutine force_gluon(this,BB,idepth,u,wl34,wl98)
!
! Calculate MD Force from gauge action.
!
  implicit none
  type(hmc_gluon),       intent(in)    :: this     ! gauge action parameters
  type(vfield_gluon_wg), intent(inout) :: BB       ! gauge pre-force field (accumulated)
  integer,               intent(in)    :: idepth   ! MD depth for recursive MD integrator
  type(vfield_gluon_wg), intent(in)    :: u        ! gauge link fields
  type(tfield_gluon_wg), intent(in)    :: wl34     ! 2-link field prepared by make_two_links
  type(tfield_gluon_wg), intent(in)    :: wl98     ! 2-link field prepared by make_two_links

  if (idepth == this%depth) then
    call force_gluon_eo(this,BB%eo(0),u%eo(0),u%eo(1),wl34%eo(0),wl34%eo(1),wl98%eo(0),wl98%eo(1))
    call force_gluon_eo(this,BB%eo(1),u%eo(1),u%eo(0),wl34%eo(1),wl34%eo(0),wl98%eo(1),wl98%eo(0))
  endif

  return
end subroutine

subroutine force_gluon_eo(this,BBe,ue,uo,wl34e,wl34o,wl98e,wl98o)
!
! MD Force from gluon action. even/odd sites only
!
!  ////NO (SF boundary condition is also included by ifdef statements.)
!
!   BB_{mu}(n) = -beta/6 cRT(0) Sum[VPL_{mu,nu}(n),nu=1,NDIM,nu=!=mu]
!                -beta/6 cRT(1) Sum[VRT_{mu,nu}(n),nu=1,NDIM,nu=!=mu],
!   force for U(n)_mu.
! 
!   VPL_{mu,nu}(n)=
! 
!     4---<---3       n       2
!     |       |       |       |
!     v       ^   +   ^       v
!     |       |       |       |
!     n       2       6---<---5
! 
!   VRT_{mu,nu}(n)=
! 
!     e---<---d
!     |       |
!     v       ^
!     |       |
!     4       3       4---<---3---<---b       7---<---4---<---3
!     |       |       |               |       |               |
!     v       ^   +   v               ^   +   v               ^
!     |       |       |               |       |               |
!     n       2       n       2--->---a       8--->---n       2
! 
!     n       2       n       2--->---a       8--->---n       2
!     |       |       |               |       |               |
!  +  ^       v   +   ^               v   +   ^               v
!     |       |       |               |       |               |
!     6       5       6---<---5---<---f       9---<---6---<---5
!     |       |
!     ^       v
!     |       |
!     i---<---h
!
  implicit none
  type(hmc_gluon),          intent(in)    :: this        ! gluon action parameters
  type(vfield_gluon_eo_wg), intent(inout) :: BBe         ! gauge pre-force (accumulated)
  type(vfield_gluon_eo_wg), intent(in)    :: ue,uo       ! gauge link field
  type(tfield_gluon_eo_wg), intent(in)    :: wl34e,wl98e ! two-link fields
  type(tfield_gluon_eo_wg), intent(in)    :: wl34o,wl98o ! two-link fields
  real(DP) :: beta_c0
  type(sfield_gluon_eo_wg), allocatable ::  vgple
  type(sfield_gluon_eo_wg), allocatable ::  w1e
  type(sfield_gluon_eo_wg), allocatable ::  w52e
  type(sfield_gluon_eo_wg), allocatable ::  w34e
  type(sfield_gluon_eo_wg), allocatable ::  w61e
#ifdef _RGGAUGE
  real(DP) :: beta_c1
  type(sfield_gluon_eo_wg), allocatable ::  vgrte
  type(sfield_gluon_eo_wg), allocatable ::  w2e,w3e,w4e
  type(vfield_gluon_eo_wg), allocatable ::  wl8e
#endif
  integer :: mu,nu,ipl
  character(CHARLEN) :: str


  if (.not. (( ue%ieo /= uo%ieo   ) .and.  &
 &           (BBe%ieo == ue%ieo   ) .and.  &
 &           (BBe%ieo == wl34e%ieo) .and.  &
 &           (BBe%ieo == wl98e%ieo) .and.  &
 &           (BBe%ieo /= uo%ieo   ) .and.  &
 &           (BBe%ieo /= wl34o%ieo) .and.  &
 &           (BBe%ieo /= wl98o%ieo)) ) then
    write(str,'("error in force_gluon_eo(even/odd)",A,I4)')__FILE__,__LINE__
    call error_stop(str)
  endif

  allocate(vgple,w1e,w52e,w34e,w61e)
  call new(vgple,ue%ieo)

#ifdef _RGGAUGE
  allocate(vgrte,w2e,w3e,w4e,wl8e)
  call new(vgrte,ue%ieo)

  call new(wl8e,ue%ieo)
  do mu=1,NDIM
    call assign_conjg_shift(wl8e%mu(mu),uo%mu(mu),-mu)
    call copy_boundary(wl8e%mu(mu))
  enddo
#ifdef _SF
!=================================================================
! SF boundary condition is broken by the previous copy_boundary
!=================================================================
  call set_sf_boundary(wl8e,conjugate=.true.)
#endif

  beta_c0 =-(this%gluon%beta/2.0_DP)*this%gluon%cRG(0)/COL
  beta_c1 =-(this%gluon%beta/2.0_DP)*this%gluon%cRG(1)/COL
#else
  beta_c0 =-(this%gluon%beta/2.0_DP)/COL
#endif

!************************************************************************
  do mu=1,NDIM

    call clear(vgple)
#ifdef _RGGAUGE
    call clear(vgrte)
#endif

    do nu=1,NDIM
      if (mu == nu) cycle
      ipl=imunu(mu,nu)

!****************************************************************
!* Force from Plaquette part of gauge action
!****************************************************************
      if (ipl > 0) then
        call             assign(w34e,wl34e%munu( ipl))
        call       assign_shift(w61e,wl98o%munu( ipl),mu)
      else
        call       assign_conjg(w34e,wl34e%munu(-ipl))
        call assign_conjg_shift(w61e,wl98o%munu(-ipl),mu)
      endif
!**********************************
!*                   
!*   4---<---3       
!*   |       |       
!*   v       ^       
!*   |       |       
!*   n       2   w341
!*
!**********************************
      call assign_conjg2_mult(w1e,w34e,ue%mu(nu))
#ifdef _SF
      if (4 == mu) then
        call accum_mult_at(w1e,this%gluon%ct,itt0=1,itt1=NTT)
      else if (4 == nu) then
        call accum_mult_at(w1e,this%gluon%ct,itt0=NTT)
      endif
#endif
      call accum_add(vgple,w1e)

!**********************************
!*                   
!*   n       2       
!*   |       |       
!*   ^       v       
!*   |       |       
!*   6---<---5   w561
!*
!**********************************
      call assign_shift(w52e,ue%mu(nu),mu,-nu)
      call assign_conjg1_mult(w1e,w52e,w61e)
#ifdef _SF
      if (4 == mu) then
        call accum_mult_at(w1e,this%gluon%ct,itt0=1,itt1=NTT)
      else if (4 == nu) then
        call accum_mult_at(w1e,this%gluon%ct,itt0=2)
      endif
#endif
      call accum_add(vgple,w1e)

#ifdef _RGGAUGE
!****************************************************************
!* Force from Rectanluar part of gauge action
!****************************************************************
!**********************************
!*                   
!*   e---<---d       
!*   |       |       
!*   v       ^       
!*   |       |       
!*   4       3   w3de4 (=> w3e)
!*           |       
!*           ^       
!*           |       
!*           2       
!*
!**********************************
      call assign_shift(w1e,uo%mu(nu),mu)
      if (ipl > 0) then
        call       assign_shift(w3e,wl34o%munu( ipl),nu)
      else
        call assign_conjg_shift(w3e,wl34o%munu(-ipl),nu)
      endif
      call        assign_mult(w2e,w1e,w3e)
      call       assign_shift(w1e,uo%mu(nu),nu)
      call assign_conjg2_mult(w3e,w2e,w1e)

!**********************************
!*                   
!*         wab34 (=> w4e)
!*
!*   4---<---3---<---b
!*                   |
!*                   ^
!*                   |
!*   n       2--->---a
!*
!**********************************
      call assign_shift(w1e,uo%mu(mu),mu)
      if (ipl > 0) then
        call       assign_shift(w4e,wl34o%munu( ipl),mu)
      else
        call assign_conjg_shift(w4e,wl34o%munu(-ipl),mu)
      endif
      call        assign_mult(w2e,w1e,w4e)
      call       assign_shift(w1e,uo%mu(mu),nu)
      call assign_conjg2_mult(w4e,w2e,w1e)

!**********************************
!*                   
!*   e---<---d       
!*   |       |       
!*   v       ^       
!*   |       |       
!*   4       3       4---<---3---<---b       4
!*     (w3e) |                       |       |
!* (   w3de4 ^   +        wab34(w4e) ^  )*   v   =  w3de41_wab341
!*           |                       |       |
!*   n       2       n       2--->---a       n
!*
!**********************************
#ifdef _SF
      if (4 == mu) then
        call accum_mult_at(w3e,this%gluon%ct_rg,itt0=1,itt1=NTT)
        call accum_mult_at(w4e,ZERO,itt0=NTT)
      else if (4 == nu) then
        call accum_mult_at(w3e,ZERO,itt0=NTT)
        call accum_mult_at(w4e,this%gluon%ct_rg,itt0=NTT)
      endif
#endif
      call         assign_add(w1e,w3e,w4e)
      call assign_conjg2_mult(w2e,w1e,ue%mu(nu))
      call          accum_add(vgrte,w2e)

!**********************************
!*                   
!*       w34781 (=>w1e)
!*
!*   7---<---4---<---3
!*   |               |
!*   v               ^
!*   |               |
!*   8--->---n       2
!*
!**********************************
      call       assign_shift(w1e,ue%mu(mu),-mu,nu)
      call assign_conjg2_mult(w2e,w34e,w1e)
      if (ipl > 0) then
        call assign_conjg_shift(w3e,wl98o%munu( ipl),nu)
      else
        call       assign_shift(w3e,wl98o%munu(-ipl),nu)
      endif
      call assign_mult(w1e,w2e,w3e)
#ifdef _SF
      if (4 == mu) then
        call accum_mult_at(w1e,ZERO,itt0=1)
      else if (4 == nu) then
        call accum_mult_at(w1e,this%gluon%ct_rg,itt0=NTT)
      endif
#endif
      call   accum_add(vgrte,w1e)

!**********************************
!*                   
!*   n       2       
!*   |               
!*   ^               
!*   |               
!*   6       5   whi61 (=>w3e)
!*   |       |       
!*   ^       v       
!*   |       |       
!*   i---<---h       
!*
!**********************************
      call assign_shift(w1e,wl8e%mu(nu),mu,-nu)
      if (ipl > 0) then
        call       assign_shift(w3e,wl98e%munu( ipl),mu,-nu)
      else
        call assign_conjg_shift(w3e,wl98e%munu(-ipl),mu,-nu)
      endif
      call  assign_mult(w2e,w1e,w3e)
      call assign_shift(w1e,uo%mu(nu),-nu)
      call  assign_mult(w3e,w2e,w1e)

!**********************************
!*
!*   8--->---n       2
!*   |                
!*   ^                
!*   |                
!*   9---<---6---<---5
!*
!*       w6981 (=>w4e)
!*                   
!**********************************
      call assign_shift(w1e,uo%mu(mu),-nu)
      if (ipl > 0) then
        call       assign(w4e,wl98e%munu( ipl))
      else
        call assign_conjg(w4e,wl98e%munu(-ipl))
      endif
      call assign_conjg1_mult(w2e,w1e,w4e)
      call       assign_shift(w1e,uo%mu(mu),-mu)
      call        assign_mult(w4e,w2e,w1e)

!**********************************
!*                   
!*         n       2       
!*         |               
!*         v
!*         |               
!*  2      6       5       8--->---n       2   
!*  |      | (w3e) |       |                   
!*  v   *( ^ whi61 v   +   ^    w6981 (w4e)   )   =  w5hi61_w56981
!*  |      |       |       |                   
!*  5      i---<---h       9---<---6---<---5  
!*
!**********************************
#ifdef _SF
      if (4 == mu) then
        call accum_mult_at(w3e,this%gluon%ct_rg,itt0=1,itt1=NTT)
        call accum_mult_at(w4e,ZERO,itt0=1)
      else if (4 == nu) then
        call accum_mult_at(w3e,ZERO,itt0=2)
        call accum_mult_at(w4e,this%gluon%ct_rg,itt0=2)
      endif
#endif
      call         assign_add(w1e,w3e,w4e)
      call assign_conjg1_mult(w2e,w52e,w1e)
      call          accum_add(vgrte,w2e)

!**********************************
!*
!*   n       2--->---a
!*   |               |
!*   ^               v
!*   |               |
!*   6---<---5---<---f
!*
!*      waf561 (=>w1e)
!*
!**********************************
      call assign_shift(w1e,ue%mu(mu),mu,-nu)
      if (ipl > 0) then
        call assign_conjg_shift(w3e,wl34e%munu( ipl),mu,-nu)
      else
        call       assign_shift(w3e,wl34e%munu(-ipl),mu,-nu)
      endif
      call assign_conjg2_mult(w2e,w3e,w1e)
      call        assign_mult(w1e,w2e,w61e)
#ifdef _SF
      if (4 == mu) then
        call accum_mult_at(w1e,ZERO,itt0=NTT)
      else if (4 == nu) then
        call accum_mult_at(w1e,this%gluon%ct_rg,itt0=2)
      endif
#endif
      call          accum_add(vgrte,w1e)

!********** end of ifdef _RGGAUGE
#endif

    enddo ! end of do nu

    call assign_mult(w1e,vgple,beta_c0)
    call   accum_add(BBe%mu(mu),w1e)
#ifdef _RGGAUGE
    call assign_mult(w1e,vgrte,beta_c1)
    call   accum_add(BBe%mu(mu),w1e)
#endif

  enddo ! end of do mu

  deallocate(vgple,w1e,w52e,w34e,w61e)
#ifdef _RGGAUGE
  deallocate(vgrte,w2e,w3e,w4e,wl8e)
#endif

  return
end subroutine

subroutine hamil_gluon(this,ifi,u,p,SG,SP,wl34,wl98)
!
! Calculate Action value for gluon action (with Plaquette + Rectangle)
! 
!  SG = beta/6 Sum[ c0   Re[Tr[1-Upl_{mu,nu}(n)]]
!                  +c1 2 Re[Tr[1-Urt_{mu,nu}(n)]] ,
!                             n=VOL, mu=1..NDIM, nu=1..NDIM, mu=!=nu]
! 
  use hmc_status_class
  implicit none
  type(hmc_gluon),        intent(inout) :: this  ! HMC gauge action
  integer,                intent(in)    :: ifi   ! 0 for initail momentum is generated, 1 for final, 2 for final
  type(vfield_gluon_wg),  intent(inout) :: u   ! gauge field
  type(vfield_gluon_wog), intent(inout) :: p   ! momentum field
  real(DP),               intent(out)   :: SG  ! gauge action value
  real(DP),               intent(out)   :: SP  ! gauge kinetic term value
  type(tfield_gluon_wg),  intent(in)    :: wl34 ! 2-links field prepared by make_two_links
  type(tfield_gluon_wg),  intent(in)    :: wl98 ! 2-links field prepared by make_two_links

  type(wloops_obj)   :: WLPe,WLPo
  real(DP)           :: SGe,SPe,SGo,SPo
  character(CHARLEN) :: str
  type(hmc_status) :: status

  if (ifi==0) call set_gaussian_noise(p) ! update p with Heatbath

  if (ifi==1) then
    call reunit_u_gluon(u)
    call copy_boundary(u)
  endif

  call hamil_gluon_eo(this,SGe,SPe,WLPe,p%eo(0),u%eo(1),wl34%eo(0),wl34%eo(1),wl98%eo(0))
  call hamil_gluon_eo(this,SGo,SPo,WLPo,p%eo(1),u%eo(0),wl34%eo(1),wl34%eo(0),wl98%eo(1))

  SG=SGe+SGo
  SP=SPe+SPo
  
  this%WLP%PLQ=WLPe%PLQ+WLPo%PLQ
  this%WLP%RCT=WLPe%RCT+WLPo%RCT

  write(str,'("@",I8,I2," Gauge: ID:",I3,"   SU:",E24.16,"   SP:",E24.16,"  PLQ:",E24.16)') &
 &  get_trajectory_number(status),ifi,this%myid,SG,SP,this%WLP%PLQ
  call print_log_action(status,str)

  return
end subroutine

subroutine hamil_gluon_eo(this,SG,SP,WLP,pe,uo,wl34e,wl34o,wl98e)
!
! Calculate Hamiltonian with Plaquette + Rectangle gauge action
! on even/odd sites only 
! 
!  SG = beta/6 Sum[ c0   Re[Tr[1-Upl_{mu,nu}(n)]]
!                  +c1 2 Re[Tr[1-Urt_{mu,nu}(n)]] ,
!                             n=VOL, mu=1..NDIM, nu=1..NDIM, mu=!=nu]
! 
  use comlib
  implicit none
  type(hmc_gluon),           intent(in)  :: this  ! gluon aciont
  real(DP),                  intent(out) :: SG    ! action value
  real(DP),                  intent(out) :: SP    ! kinetica term value
  type(wloops_obj),          intent(out) :: WLP   ! wilson loops value
  type(vfield_gluon_eo_wog), intent(in)  :: pe    ! gauge momentum on even/odd sites
  type(vfield_gluon_eo_wg),  intent(in)  :: uo    ! gauge field on odd/even sites
  type(tfield_gluon_eo_wg),  intent(in)  :: wl34e ! 2-links field on even/odd sites
  type(tfield_gluon_eo_wg),  intent(in)  :: wl34o ! 2-links field on odd/even sites
  type(tfield_gluon_eo_wg),  intent(in)  :: wl98e ! 2-links field on even/odd sites

  integer  :: mu,nu,ipl
  real(DP) :: dof
  real(DP) :: rtrace
  real(DP) :: beta_c0
  real(DP) :: rtmp0,SPL
  type(sfield_gluon_eo_wg), allocatable :: w1e,w2e
  type(sfield_gluon_eo_wg), allocatable :: w12e
#ifdef _RGGAUGE
  real(DP) :: beta_c1
  real(DP) :: rtmp1,SRC
  type(sfield_gluon_eo_wg), allocatable :: w3e
#endif

#ifdef _RGGAUGE
  beta_c0=(this%gluon%beta/2.0_DP)*this%gluon%cRG(0)/COL
  beta_c1=(this%gluon%beta       )*this%gluon%cRG(1)/COL
#else
  beta_c0=(this%gluon%beta/2.0_DP)/COL
#endif

  allocate(w1e,w2e,w12e)
  SP    = 0.0_DP
  rtmp0 = 0.0_DP
  SPL   = 0.0_DP
#ifdef _RGGAUGE
  allocate(w3e)
  rtmp1 = 0.0_DP
  SRC   = 0.0_DP
#endif

  do mu=1,NDIM
    do nu=1,NDIM
      if (mu == nu) cycle
      ipl=imunu(mu,nu)

!*****************************************************
!*   
!*  n+nu          n+nu           n+2nu
!*   4---<---3     4              e---<---d
!*           |     |                      |
!*     w34   ^     v  w12            wde  ^
!*           |     |                      |
!*   1       2     1--->---2      4       3
!*   n      n+mu   n      n+mu   n+nu    n+mu+nu
!*
!*****************************************************
      if (ipl > 0) then
        call assign_conjg_shift(w12e,wl98e%munu( ipl),mu,nu)
        call assign(      w2e,wl34e%munu( ipl))
      else
        call assign_shift(w12e,wl98e%munu(-ipl),mu,nu)
        call assign_conjg(w2e,wl34e%munu(-ipl))
      endif

!***************************
!* 
!*  n+nu
!*   4---<---3
!*   |       |
!*   v       ^  w1234
!*   |       |
!*   1--->---2
!*   n      n+mu
!*
!***************************
!* trace plaquette and action
!***********************************
      call assign_mult(w1e,w12e,w2e)
#ifdef _SF
      if ((4 /= mu).and.(4 /= nu)) then
        !===========================================================
        ! spatial plaquettes on itt=1 boundary have zero contribution
        !===========================================================
        call accum_mult_at(w1e,ZERO,itt0=1)
      endif
#endif
      rtrace=real(trace(w1e))
      rtmp0=rtmp0+rtrace     ! plaquette average
#ifdef _SF
      if ((4 == mu).or.(4 == nu)) then
        !===========================================================
        ! temp-space plaquettes touching itt=1 and NTT+1 
        ! boundary have a spatial weight to the action
        !
        !  t=2   x---<---x       t=NTT+1 --x---<---x--
        !        |       |                 |       |
        !        v  ct   ^                 v  ct   ^      nu/mu
        !        |       |                 |       |       ^
        !  t=1 --n--->---x--     t=NTT     n--->---x       o-> mu/nu
        !
        !===========================================================
        call accum_mult_at(w1e,this%gluon%ct,itt0=1,itt1=NTT)
      endif
      rtrace=real(trace(w1e))
#endif
      SPL=SPL+beta_c0*(-rtrace)

#ifdef _RGGAUGE
!********************** RG-Rectangle part
!***************************
!*   
!* n+2nu
!*   e---<---d
!*           |
!*     w3de  ^
!*           |
!*   4       3
!*           |
!*           ^
!*           |
!*   1       2
!*   n      n+mu
!*
!***************************
      call assign_shift(w1e,uo%mu(nu),mu)
      if (ipl > 0) then
        call       assign_shift(w2e,wl34o%munu( ipl),nu)
      else
        call assign_conjg_shift(w2e,wl34o%munu(-ipl),nu)
      endif
      call assign_mult(w3e,w1e,w2e)

!***************************
!*
!*   e       d
!*   |        
!*   v        
!*   |        
!*   4 w412  3
!*   |        
!*   v        
!*   |        
!*   1--->---2
!*   n      n+mu
!*
!***************************
      call       assign_shift(w1e,uo%mu(nu),nu)
      call assign_conjg1_mult(w2e,w1e,w12e)

!***************************
!*
!* n+2nu
!*   e---<---d
!*   |       |
!*   v       ^
!*   |       |
!*   4       3  w4123de
!*   |       |
!*   v       ^
!*   |       |
!*   1--->---2
!*   n      
!*
!***************************
!******* trace RG-Rectangle and action
      call assign_mult(w1e,w2e,w3e)

#ifdef _SF
      if ( 4 == nu ) then
        !==========================================================
        ! temp-space rects extending t=NTT+1 boundary have zero weight
        !
        !  t=NTT+2    n---<---x
        !             |       |
        !             v       ^
        !             |       |
        !  t=NTT+1 ---x-------x---
        !             |       |
        !             v       ^    nu
        !             |       |    ^
        !  t=NTT      n--->---x    |->mu
        !
        !==========================================================
        call accum_mult_at(w1e,ZERO,itt0=NTT)
      else if ((4 /= mu).and.(4 /= nu)) then
        !==========================================================
        ! spatial rects on t=1 boundary have zero weight
        !==========================================================
        call accum_mult_at(w1e,ZERO,itt0=1)
      endif
#endif
      rtrace=real(trace(w1e))
      rtmp1=rtmp1+rtrace
#ifdef _SF
      if ( 4 == mu ) then
        !==========================================================
        ! temp-space rects touching t=NTT+1 boundary in parallel
        ! have a special weight to the action.
        !
        !  t=NTT+1 ---x--->---x--->---x----------
        !             |               |
        !             ^    ct_rg      v    mu
        !             |               |    ^
        !  t=NTT      n---<---x---<---x    |->nu
        !
        !==========================================================
        call accum_mult_at(w1e,this%gluon%ct_rg,itt0=1,itt1=NTT)
      endif
      rtrace=real(trace(w1e))
#endif
      SRC=SRC+beta_c1*(-rtrace)
!******** end of ifdef _RGGAUGE
#endif
    enddo  ! end of do nu

!**************************************
!*  tr[P^2] = sum[|P(a,b)|^2,{a=1,3},{b=1,3}]
!**************************************
    SP=SP+abs2(pe%mu(mu))

  enddo  ! end of do mu

!***************
!* global sum
!***************

  SP=SP/2.0_DP

  deallocate(w1e,w2e,w12e)

#ifdef _SF
  dof = 0.0_DP
#ifndef _RGGAUGE
  WLP%PLQ = rtmp0/(NTX*NTY*NTZ*(NDIM*(NTT-1)+2))/(NDIM-1)/COL
  WLP%RCT = 0.0_DP
  SG= beta_c0*dof+SPL
#else
  deallocate(w3e)
  SG=(beta_c0+beta_c1)*dof+(SPL+SRC)
  WLP%PLQ = rtmp0/(NTX*NTY*NTZ*(NDIM*(NTT-1)+2))/(NDIM-1)/COL
  WLP%RCT = rtmp1/(NTX*NTY*NTZ*(NDIM*(NTT-1)+2))/(NDIM-1)/COL
#endif
#else
  dof=real(NTH*NTZ*NTY*NTX*COL*NDIM*(NDIM-1),kind=DP)
#ifndef _RGGAUGE
  WLP%PLQ = rtmp0/(NTX*NTY*NTZ*NTT)/(NDIM-1)/NDIM/COL
  WLP%RCT = 0.0_DP
  SG= beta_c0*dof+SPL
#else
  deallocate(w3e)
  SG=(beta_c0+beta_c1)*dof+(SPL+SRC)
  WLP%PLQ = rtmp0/(NTX*NTY*NTZ*NTT)/(NDIM-1)/NDIM/COL
  WLP%RCT = rtmp1/(NTX*NTY*NTZ*NTT)/(NDIM-1)/NDIM/COL
#endif
#endif

  return
end subroutine

subroutine reunit_u_gluon(u,reunit_log,hmc_reunit_log)
!
! Re-unitarize gauge field including ghost sites
!
!  u : gauge link
!
  use comlib
  implicit none
  type(vfield_gluon_wg),       intent(inout) :: u
  type(logfile),     optional, intent(inout) :: reunit_log
  type(hmc_logfile), optional, intent(inout) :: hmc_reunit_log
  if (present(reunit_log)) then
    call reunit_u_gluon_eo(u%eo(0),reunit_log=reunit_log)
    call reunit_u_gluon_eo(u%eo(1),reunit_log=reunit_log)
  else if (present(hmc_reunit_log)) then
    call reunit_u_gluon_eo(u%eo(0),hmc_reunit_log=hmc_reunit_log)
    call reunit_u_gluon_eo(u%eo(1),hmc_reunit_log=hmc_reunit_log)
  else
    call reunit_u_gluon_eo(u%eo(0))
    call reunit_u_gluon_eo(u%eo(1))
  endif
  return
end subroutine

subroutine reunit_u_gluon_eo(ue, reunit_log, hmc_reunit_log)
!
! Re-unitarize gauge links including ghost sites on even/odd sites
!
  use comlib
  implicit none
  type(vfield_gluon_eo_wg),    intent(inout) :: ue
  type(logfile),     optional, intent(inout) :: reunit_log
  type(hmc_logfile), optional, intent(inout) :: hmc_reunit_log

  integer :: ic,jc,mu,ix,iy,iz,itb
  complex(DP) ::  un(COL,COL),zunit(COL,COL),ztmp
  real(DP), parameter :: gamma = 10.0e0_DP
  real(DP) :: rnorm
  integer  :: iunitz,ipe,inum(0:NPU-1)
  character(CHARLEN) :: str

  iunitz=0
  do mu=1,NDIM

!$OMP PARALLEL DO &
!$OMP COLLAPSE(3) &
!$OMP PRIVATE(ix,iy,iz,itb,ic,jc,zunit,ztmp,un,rnorm) REDUCTION(+:iunitz)
  do ix=0,NX1
  do iy=0,NY1
  do iz=0,NZ1

    if ((ix ==   0 .and. iy ==   0 .and. iz ==   0) .or. &
 &      (ix == NX1 .and. iy ==   0 .and. iz ==   0) .or. &
 &      (ix ==   0 .and. iy == NY1 .and. iz ==   0) .or. &
 &      (ix == NX1 .and. iy == NY1 .and. iz ==   0) .or. &
 &      (ix ==   0 .and. iy ==   0 .and. iz == NZ1) .or. &
 &      (ix == NX1 .and. iy ==   0 .and. iz == NZ1) .or. &
 &      (ix ==   0 .and. iy == NY1 .and. iz == NZ1) .or. &
 &      (ix == NX1 .and. iy == NY1 .and. iz == NZ1) ) cycle

  do itb=0,NTH
    un(:,:)=ue%mu(mu)%s(itb,iz,iy,ix)%u(:,:)

    !---------------------
    ! check unitarity
    !      U*U^dag = 1?
    !---------------------
    do jc=1,COL
    do ic=1,COL
      zunit(ic,jc) = un(ic,1)*CONJG(un(jc,1))  &
 &                  +un(ic,2)*CONJG(un(jc,2))  &
 &                  +un(ic,3)*CONJG(un(jc,3))
    enddo
    enddo

    rnorm = (real(zunit(1,1))-1.0_DP)**2  &
 &         + real(zunit(1,2))**2  &
 &         + real(zunit(1,3))**2  &
 &         + real(zunit(2,1))**2  &
 &         +(real(zunit(2,2))-1.0_DP)**2  &
 &         + real(zunit(2,3))**2  &
 &         + real(zunit(3,1))**2  &
 &         + real(zunit(3,2))**2  &
 &         +(real(zunit(3,3))-1.0_DP)**2  &
 &         +aimag(zunit(1,1))**2  &
 &         +aimag(zunit(1,2))**2  &
 &         +aimag(zunit(1,3))**2  &
 &         +aimag(zunit(2,1))**2  &
 &         +aimag(zunit(2,2))**2  &
 &         +aimag(zunit(2,3))**2  &
 &         +aimag(zunit(3,1))**2  &
 &         +aimag(zunit(3,2))**2  &
 &         +aimag(zunit(3,3))**2
    if (rnorm > (COL*COL*2*EPSILON(1.0_DP)*gamma)**2) iunitz=iunitz+1

    !----------------------------------
    ! re-orthogonalzation
    ! normalize first column u(1)
    !----------------------------------
    rnorm = REAL(un(1,1))**2 + AIMAG(un(1,1))**2  &
 &         +REAL(un(2,1))**2 + AIMAG(un(2,1))**2  &
 &         +REAL(un(3,1))**2 + AIMAG(un(3,1))**2
    rnorm = 1.0_DP/SQRT(rnorm)
    un(:,1) = un(:,1)*rnorm

    !----------------------------------
    ! orthgonalize second column u(2)
    ! against u(1)
    !----------------------------------
    ztmp = CONJG(un(1,1))*un(1,2) &
 &        +CONJG(un(2,1))*un(2,2) &
 &        +CONJG(un(3,1))*un(3,2)
    un(:,2) = un(:,2) - un(:,1)*ztmp
    rnorm = REAL(un(1,2))**2 + AIMAG(un(1,2))**2  &
 &         +REAL(un(2,2))**2 + AIMAG(un(2,2))**2  &
 &         +REAL(un(3,2))**2 + AIMAG(un(3,2))**2
    rnorm = 1.0_DP/SQRT(rnorm)
    un(:,2) = un(:,2)*rnorm

    !----------------------------------
    ! reconstruct third comlumn u(3)
    ! with u(3) = (u(1)xu(2))^*
    ! using special unitarity SU(3)
    !----------------------------------
    un(1,3) = CONJG(un(2,1)*un(3,2)-un(3,1)*un(2,2))
    un(2,3) = CONJG(un(3,1)*un(1,2)-un(1,1)*un(3,2))
    un(3,3) = CONJG(un(1,1)*un(2,2)-un(2,1)*un(1,2))

    ue%mu(mu)%s(itb,iz,iy,ix)%u(:,:) = un(:,:)

  enddo
  enddo
  enddo
  enddo ! end of do ix,iy,iz,itb
  enddo ! end of do mu

!********** collect the number of unitariy violating link variables
  inum(:) = 0
  inum(nodeid)=iunitz
#ifndef _singlePU
  call comlib_sumcast(inum)
#endif
  if (nodeid == 0) then
    do ipe=0,NPU-1
      if (inum(ipe) /= 0) then
        write(str,'("nodeid :",I3," BAD LINK # ",I9)')ipe,inum(ipe)

        if (present(reunit_log)) then

          call print(reunit_log,str)

        else if (present(hmc_reunit_log)) then

          call print(hmc_reunit_log,str)

        else

          call print(m_reunit_log,str)

        endif
      endif
    enddo
  endif

  return
end subroutine

function delta_p(p0,p1) result(diff)
!
! compute difference norm of two momentum fields (for reversibility check)
!
!  delta_p = (|p0 - p1|^2/(COL^2*NDIM*NTX*NTY*NTZ*NTT))^(1/2)
!       p0 : momentum 1
!       p1 : momentum 2
!
  use comlib
  implicit none
  real(DP) :: diff
  type(vfield_gluon_wog), intent(in) :: p0,p1

  integer :: ix,iy,iz,itb,ieoxyz,ieo,ic,jc,mu

  diff = 0.0_DP
  do ieo=0,1
  do mu=1,NDIM
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    do itb=1,NTH
      do jc=1,COL
      do ic=1,COL
        diff = diff + real( p0%eo(ieo)%mu(mu)%s(itb,iz,iy,ix)%u(ic,jc)      &
 &                         -p0%eo(ieo)%mu(mu)%s(itb,iz,iy,ix)%u(ic,jc))**2  &
 &                 + aimag( p0%eo(ieo)%mu(mu)%s(itb,iz,iy,ix)%u(ic,jc)      &
 &                         -p1%eo(ieo)%mu(mu)%s(itb,iz,iy,ix)%u(ic,jc))**2
      enddo
      enddo
    enddo
  enddo
  enddo
  enddo ! end of do ix,iy,iz
  enddo ! end of do mu
  enddo
#ifndef _singlePU
  call comlib_sumcast(diff)
#endif

  diff = sqrt(diff/(COL**2*NDIM)/(NTT*NTZ*NTY*NTX))

  return
end function

function delta_u(u0,u1) result(diff)
!
! compute diferrence norm of two gauge fields (for reversibility check)
!
!  delta_u = (|u0 - u1|^2/(COL^2*NDIM*NTX*NTY*NTZ*NTT))^(1/2)
!       u0 : link 1
!       u1 : link 2
!
  use comlib
  implicit none
  real(DP) :: diff
  type(vfield_gluon_wg), intent(in) :: u0,u1

  integer :: ix,iy,iz,itb,ieoxyz,ieo,ic,jc,mu

  diff = 0.0_DP
  do ieo=0,1
  do mu=1,NDIM
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+iz+iy+ix+ieo,2)
    do itb=1-ieoxyz,NTH-ieoxyz
      do jc=1,COL
      do ic=1,COL
        diff = diff + real( u0%eo(ieo)%mu(mu)%s(itb,iz,iy,ix)%u(ic,jc)      &
 &                         -u0%eo(ieo)%mu(mu)%s(itb,iz,iy,ix)%u(ic,jc))**2  &
 &                 + aimag( u0%eo(ieo)%mu(mu)%s(itb,iz,iy,ix)%u(ic,jc)      &
 &                         -u1%eo(ieo)%mu(mu)%s(itb,iz,iy,ix)%u(ic,jc))**2
      enddo
      enddo
    enddo
  enddo
  enddo
  enddo ! end of do ix,iy,iz
  enddo ! end of do mu
  enddo
#ifndef _singlePU
  call comlib_sumcast(diff)
#endif

  diff = sqrt(diff/(COL**2*NDIM)/(NTT*NTZ*NTY*NTX))

  return
end function

end module
