#include <config.h>

module hmc_quark_class
!
!= Defines HMC QCD quark actions
!
!== Version
!
! $Id: hmc_quark_class.F90,v 1.18 2011/10/18 07:18:00 ishikawa Exp $
!
  use lattice_class
  use error_class
  use hmc_status_class
  use field_gauge_class
  use field_fermion_class
#ifdef _CLOVER
  use quark_clover_class
#else
  use quark_wilson_class
#endif
  use hmc_truncd_overlap_class
  use hmc_std_quark_wilson_class
  use hmc_blk_quark_wilson_class
  use hmc_poly_quark_wilson_class
  use hmc_mp_quark_wilson_class
  implicit none
  private

  public :: copy_fq_time
  public :: print_copy_fq_statistics
  public :: mult_iter
  public :: hmc_quark
  public :: metropolis
  public :: reweighting  ! determinant reweighting
  public :: new
  public :: delete
  public :: read
  public :: print
  public :: print_statistics
  public :: hamil
  public :: force
  public :: save_config
  public :: set_two_links
  public :: remove_two_links

  integer, save :: m_num_quark=0   ! this counts total number of quark action/algorithm
  type(tfield_gluon_wg), private, pointer :: m_wl34 => NULL()
  type(tfield_gluon_wg), private, pointer :: m_wl98 => NULL()

  type hmc_quark
  !
  ! HMC QCD quark action
  !
  ! This can have several actions.
  !
    character(CHARLEN), public :: file_name
    type(hmc_poly_quark_wilson),  private, pointer ::  poly_hmc => NULL()  ! algorithm = PHMC_ALGORITHM = 0
    type(hmc_std_quark_wilson),   private, pointer ::   std_hmc => NULL()  ! algorithm = HMC_ALGORITHM = 1
    type(hmc_mp_quark_wilson),    private, pointer ::    mp_hmc => NULL()  ! algorithm = MPHMC_ALGORITHM = 2
    type(hmc_blk_quark_wilson),   private, pointer ::   blk_hmc => NULL()  ! algorithm = HMC_BLK_ALGORITHM = 3
    type(hmc_truncd_dw_overlap),  private, pointer :: tdwov_hmc => NULL()  ! algorithm = HMC_ALGORITHM_TRUNCD_DW_OVERLAP = 10
    integer, private :: algorithm = 0
    integer, private :: myid      = 0
    integer, private :: irun      = 0
    integer, private :: idummy    = 0
  end type

  interface new
  !
  ! Initialize/create HMC QCD quark action
  !
    module procedure new_quark_array
  end interface

  interface delete
  !
  ! delete HMC QCD quark action
  !
    module procedure delete_quark_array
  end interface

  interface read
  !
  ! read HMC QCD quark action parameter form file
  !
    module procedure read_quark_array
  end interface

  interface save_config
  !
  ! save HMC QCD quark action parameter to field config file
  !
    module procedure save_config_quark_array
  end interface

  interface print
  !
  ! print out HMC QCD quark action paramters on display
  !
    module procedure print_quark_array
  end interface

  interface print_statistics
  !
  ! print out HMC QCD quark action job statistics on display
  !
    module procedure print_stat_quark_array
  end interface

  interface hamil
  !
  ! compute HMC QCD quark action value
  !
    module procedure hamil_quark_array
  end interface

  interface force
  !
  ! compute MD force for HMC QCD quark action
  !
    module procedure force_quark_array
  end interface

  interface metropolis
  !
  ! do Metropolis test for HMC QCD quark action if it contais Global Metropolis test
  !
    module procedure metropolis_quark_array
  end interface

  interface reweighting
  !
  ! compute reweighting factor for quark determinant
  !
    module procedure reweighting_quark_array
  end interface

contains

subroutine new_quark_array(this)
!
! Initialize/create HMC quark action array
!
  implicit none
  type(hmc_quark),  intent(inout)      :: this(:)
  integer :: iqk,nqk
  nqk = SIZE(this)
  do iqk=1,nqk
    call new_quark(this(iqk),iqk)
  enddo
  return
end subroutine

subroutine new_quark(this,id)
!
! Initialize/create HMC quark action
!
! This class mimics dynamic polymorphism to quark type.
!
  implicit none
  type(hmc_quark),  intent(inout) :: this
  integer,          intent(in)    :: id
  m_num_quark = m_num_quark+1
  this%myid   = id
  this%poly_hmc => NULL()
  this%std_hmc  => NULL()
  this%mp_hmc   => NULL()
  return
end subroutine

subroutine delete_quark_array(this)
!
! delete HMC quark actions (array)
!
  implicit none
  type(hmc_quark), intent(inout) :: this(:)
  integer :: iqk,nqk
  nqk = SIZE(this)
  do iqk=1,nqk
    call delete_quark(this(iqk))
  enddo
  return
end subroutine

subroutine delete_quark(this)
!
! delete HMC quark action
!
  implicit none
  type(hmc_quark), intent(inout) :: this

  select case(this%algorithm)
  case(PHMC_ALGORITHM)
    call delete(this%poly_hmc)
    deallocate(this%poly_hmc)
  case(HMC_ALGORITHM)
    call delete(this%std_hmc)
    deallocate(this%std_hmc)
  case(MPHMC_ALGORITHM)
    call delete(this%mp_hmc)
    deallocate(this%mp_hmc)
  case(HMC_BLK_ALGORITHM)
    call delete(this%blk_hmc)
    deallocate(this%blk_hmc)
  case(HMC_ALGORITHM_TRUNCD_DW_OVERLAP)
    call delete(this%tdwov_hmc)
    deallocate(this%tdwov_hmc)
  case default
    call error_stop(" Error in quark param file (algorithm)")
  end select
  m_num_quark = m_num_quark - 1

  return
end subroutine

subroutine read_quark_array(this)
!
! Read HMC quark action parameter from formatted file (array)
!
  use comlib
  implicit none
  type(hmc_quark), intent(inout) :: this(:)
  integer :: iqk,nqk
  nqk = SIZE(this)
  do iqk=1,nqk
    call read_quark(this(iqk))
  enddo
  return
end subroutine

subroutine read_quark(this)
!
! Read HMC quark action parameter from formatted file
!
!  In this routine quark algorithm/action type is dynamically (at run time)
!  determined by the value of the input parameter file. This mimics the dynamic
!  polymorphism to the quark algorithm/action type.
!
  use comlib
  use file_tools_class
  use print_status_class
  implicit none
  type(hmc_quark), intent(inout) :: this

  character(len=CHARLEN) :: char
  integer :: iout

  if (nodeid == 0) then
    write(*,'(80("-"))')
  endif

  iout = search_free_file_unit()

  open(iout,file=this%file_name,status='unknown',form='formatted')

  read(iout,*)this%algorithm   ! read algorithm or quark type
  read(iout,*)

  !
  ! this section mimics dynamic polymorphism
  !
  select case(this%algorithm)
  case(PHMC_ALGORITHM)

    allocate(this%poly_hmc)
    call new(this%poly_hmc,this%myid)
    call read(this%poly_hmc,iout)

  case(HMC_ALGORITHM)

    allocate(this%std_hmc)
    call new(this%std_hmc,this%myid)
    call read(this%std_hmc,iout)

  case(MPHMC_ALGORITHM)

    allocate(this%mp_hmc)
    call new(this%mp_hmc,this%myid)
    call read(this%mp_hmc,iout)

  case(HMC_BLK_ALGORITHM)

    allocate(this%blk_hmc)
    call new(this%blk_hmc,this%myid)
    call read(this%blk_hmc,iout)

  case(HMC_ALGORITHM_TRUNCD_DW_OVERLAP)

    allocate(this%tdwov_hmc)
    call new(this%tdwov_hmc,this%myid)
    call read(this%tdwov_hmc,iout)

  case default
    call error_stop(" Error in quark param file (algorithm)")
  end select
  close(iout)

  write(char,'(1X,A," for quark #",I3," : read OK ",I3)')  &
                          &  TRIM(this%file_name),this%myid,nodeid
  call print_status(char)

  return
end subroutine

subroutine save_config_quark_array(this,iout)
!
! Save HMC quark action parameter on the measurement config file in unformatted form.
!
  implicit none
  type(hmc_quark), intent(in) :: this(:)
  integer, intent(in) :: iout
  integer :: iqk,nqk
  nqk = SIZE(this)
  do iqk=1,nqk
    call save_config_quark(this(iqk),iout)
  enddo
  return
end subroutine

subroutine save_config_quark(this,iout)
!
! Save HMC quark action parameter on the measurement config file in unformatted form.
!
  implicit none
  type(hmc_quark), intent(in) :: this
  integer, intent(in) :: iout

  select case(this%algorithm)
  case(PHMC_ALGORITHM)
    call save_config(this%poly_hmc,iout)
  case(HMC_ALGORITHM)
    call save_config(this%std_hmc,iout)
  case(MPHMC_ALGORITHM)
    call save_config(this%mp_hmc,iout)
  case(HMC_BLK_ALGORITHM)
    call save_config(this%blk_hmc,iout)
  case(HMC_ALGORITHM_TRUNCD_DW_OVERLAP)
    call save_config(this%tdwov_hmc,iout)
  end select

  return
end subroutine

subroutine print_quark_array(this)
!
! print out HMC QCD quark action parameters on display
!
  implicit none
  type(hmc_quark), intent(inout) :: this(:)
  integer :: iqk,nqk
  nqk = SIZE(this)
  do iqk=1,nqk
    call print_quark(this(iqk))
  enddo
  return
end subroutine

subroutine print_quark(this)
!
! print out HMC QCD quark action parameters on display
!
  implicit none
  type(hmc_quark), intent(inout) :: this
  select case (this%algorithm)
  case (PHMC_ALGORITHM)
    call print(this%poly_hmc)
  case (HMC_ALGORITHM)
    call print(this%std_hmc)
  case (MPHMC_ALGORITHM)
    call print(this%mp_hmc)
  case (HMC_BLK_ALGORITHM)
    call print(this%blk_hmc)
  case(HMC_ALGORITHM_TRUNCD_DW_OVERLAP)
    call print(this%tdwov_hmc)
  end select
  return
end subroutine

subroutine print_stat_quark_array(this)
!
! print out HMC QCD quark action job statistics on display
!
  implicit none
  type(hmc_quark), intent(inout) :: this(:)
  integer :: iqk,nqk
  nqk = SIZE(this)
  do iqk=1,nqk
    call print_stat_quark(this(iqk))
  enddo
  return
end subroutine

subroutine print_stat_quark(this)
!
! print out HMC QCD quark action job statistics on display
!
  implicit none
  type(hmc_quark), intent(inout) :: this
  select case (this%algorithm)
  case (PHMC_ALGORITHM)
    call print_statistics(this%poly_hmc)
  case (HMC_ALGORITHM)
    call print_statistics(this%std_hmc)
  case (MPHMC_ALGORITHM)
    call print_statistics(this%mp_hmc)
  case (HMC_BLK_ALGORITHM)
    call print_statistics(this%blk_hmc)
  case(HMC_ALGORITHM_TRUNCD_DW_OVERLAP)
    call print_statistics(this%tdwov_hmc)
  end select
  return
end subroutine

subroutine set_two_links(this,wl34,wl98)
  implicit none
  type(hmc_quark),       intent(inout) :: this(:)  ! quark action (array)
  type(tfield_gluon_wg), intent(in), target  :: wl34,wl98 ! 2-links fields prepared by make_two_links (need only for clover quarks)
  if (.not.associated(m_wl34)) m_wl34 => wl34
  if (.not.associated(m_wl98)) m_wl98 => wl98
  return
end subroutine

subroutine remove_two_links(this)
  implicit none
  type(hmc_quark),       intent(inout) :: this(:)  ! quark action (array)
  if (associated(m_wl34)) m_wl34 => NULL()
  if (associated(m_wl98)) m_wl98 => NULL()
  return
end subroutine

subroutine hamil_quark_array(this,ifi,u,SQ)
!
! compute HMC QCD quark Hamiltonian/Action value
! 
  implicit none
  type(hmc_quark),       intent(inout) :: this(:)  ! quark action (array)
  integer,               intent(in)    :: ifi      ! target switch: 0 for initial Hamiltonian, 1 for final, 2 initial
  type(vfield_gluon_wg), intent(in)    :: u        ! gauge field
  real(DP),              intent(out)   :: SQ(:)    ! Hamiltonian/Action values (array)

  integer :: iqk,nqk

  SQ(:) = 0.0_DP

  nqk   = SIZE(this)

#ifdef _CLOVER
  call make_clover_leaf(u,m_wl34,m_wl98)
#endif

  do iqk=1,nqk
    select case (this(iqk)%algorithm)
    case (PHMC_ALGORITHM)  ! PHMC
      call hamil(this(iqk)%poly_hmc,ifi,u,SQ(iqk))
    case (HMC_ALGORITHM) ! HMC
      call hamil(this(iqk)%std_hmc,ifi,u,SQ(iqk))
    case (MPHMC_ALGORITHM) ! MPHMC
      call hamil(this(iqk)%mp_hmc,ifi,u,SQ(iqk))
    case (HMC_BLK_ALGORITHM) ! HMCBLK
      call hamil(this(iqk)%blk_hmc,ifi,u,SQ(iqk))
    case(HMC_ALGORITHM_TRUNCD_DW_OVERLAP)
      call hamil(this(iqk)%tdwov_hmc,ifi,u,SQ(iqk))
    end select  ! end of quark algorithm
  enddo

#ifdef _CLOVER
  call delete_clover_leaf()
#endif

  return
end subroutine

subroutine force_quark_array(this,BB,idepth,u)
!
! compute MD Force from HMC QCD quark actions.
! 
  implicit none
  type(hmc_quark),       intent(inout) :: this(:)  ! quark actions (array)
  type(vfield_gluon_wg), intent(inout) :: BB       ! MD pre-force  (accumulated)
  integer,               intent(in)    :: idepth   ! MD depth for recursive MD integrator
  type(vfield_gluon_wg), intent(in)    :: u        ! gauge field
#ifdef _CLOVER
  type(tfield_gluon_wg), allocatable, target :: XX       ! clover term contribution for MD force (need only for clover quarks)
#endif

  integer :: iqk,nqk

  nqk = SIZE(this)

#ifdef _CLOVER
  !=================================
  ! Clear clover term contribution
  !=================================
  allocate(XX)
  call new(XX)
  call clear(XX)

  !=================================
  ! Calc clover leaf  F_{mu,nu}
  !=================================
  call make_clover_leaf(u,m_wl34,m_wl98)
#endif

  do iqk=1,nqk
    select case (this(iqk)%algorithm)
    case (PHMC_ALGORITHM)  ! PHMC

#ifdef _CLOVER
      this(iqk)%poly_hmc%XX => XX
#endif
      call force(this(iqk)%poly_hmc,BB,idepth,u)

    case (HMC_ALGORITHM) ! HMC

#ifdef _CLOVER
      this(iqk)%std_hmc%XX => XX
#endif
      call force(this(iqk)%std_hmc,BB,idepth,u)

    case (MPHMC_ALGORITHM) ! MPHMC

#ifdef _CLOVER
      this(iqk)%mp_hmc%XX => XX
#endif
      call force(this(iqk)%mp_hmc,BB,idepth,u)

    case (HMC_BLK_ALGORITHM) ! HMCBLK

#ifdef _CLOVER
      this(iqk)%blk_hmc%XX => XX
#endif
      call force(this(iqk)%blk_hmc,BB,idepth,u)

    case(HMC_ALGORITHM_TRUNCD_DW_OVERLAP)

      call force(this(iqk)%tdwov_hmc,BB,idepth,u)

    end select
  enddo ! end of do iqk

#ifdef _CLOVER
  call delete_clover_leaf()
  call force_clover_terms(BB,XX,u,m_wl34,m_wl98)
  deallocate(XX)
#endif

  return
end subroutine

subroutine metropolis_quark_array(this,u0,u1,ihit)
!
! Do Metropolis test for HMC QCD quark actions (array)
!
! This routine do Metropoils test when the quark action array contains 
! the actions with Global Metropolis test.
!
  implicit none
  type(hmc_quark),       intent(inout) :: this(:)  ! HMC QCD quark actions
  type(vfield_gluon_wg), intent(in)    :: u0       ! MD Initial configuration
  type(vfield_gluon_wg), intent(in)    :: u1       ! MD final configuration
  integer,               intent(inout) :: ihit     ! 1 for accept , 0 for reject
  integer :: iqk,nqk
  
  if (ihit == 0) return

  nqk = SIZE(this)
  do iqk=1,nqk
    select case(this(iqk)%algorithm)
    case(PHMC_ALGORITHM)
      if (has_global_metropolis_test(this(iqk)%poly_hmc)) then
        call metropolis(this(iqk)%poly_hmc,u1,u0,ihit)
      endif
    case(HMC_ALGORITHM)
      if (has_global_metropolis_test(this(iqk)%std_hmc)) then
        call metropolis(this(iqk)%std_hmc,u1,u0,ihit)
      endif
    case(MPHMC_ALGORITHM)
      if (has_global_metropolis_test(this(iqk)%mp_hmc)) then
        call metropolis(this(iqk)%mp_hmc,u1,u0,ihit)
      endif
    case(HMC_BLK_ALGORITHM)
      if (has_global_metropolis_test(this(iqk)%blk_hmc)) then
        call metropolis(this(iqk)%blk_hmc,u1,u0,ihit)
      endif
    case(HMC_ALGORITHM_TRUNCD_DW_OVERLAP)
      if (has_global_metropolis_test(this(iqk)%tdwov_hmc)) then
        call metropolis(this(iqk)%tdwov_hmc,u1,u0,ihit)
      endif
    case default
      call error_stop(" Error in quark param file (algorithm)")
    end select
    if (ihit == 0) return
  enddo

  return
end subroutine

subroutine reweighting_quark_array(this,u)
!
! Compute Reweighting factor for quark determinant
!
! This routine do Metropoils test when the quark action array contains 
! the actions with Global Metropolis test.
!
  implicit none
  type(hmc_quark),       intent(inout) :: this(:)  ! HMC QCD quark actions
  type(vfield_gluon_wg), intent(in)    :: u
  integer :: iqk,nqk
  
  nqk = SIZE(this)
  do iqk=1,nqk
    select case(this(iqk)%algorithm)
    case(PHMC_ALGORITHM)
      if (has_reweighting(this(iqk)%poly_hmc)) then
        call reweighting(this(iqk)%poly_hmc,u)
      endif
    case(HMC_ALGORITHM)
      if (has_reweighting(this(iqk)%std_hmc)) then
        call reweighting(this(iqk)%std_hmc,u)
      endif
    case(MPHMC_ALGORITHM)
      if (has_reweighting(this(iqk)%mp_hmc)) then
        call reweighting(this(iqk)%mp_hmc,u)
      endif
    case(HMC_BLK_ALGORITHM)
      if (has_reweighting(this(iqk)%blk_hmc)) then
        call reweighting(this(iqk)%blk_hmc,u)
      endif
    case(HMC_ALGORITHM_TRUNCD_DW_OVERLAP)
      if (has_reweighting(this(iqk)%tdwov_hmc)) then
        call reweighting(this(iqk)%tdwov_hmc,u)
      endif
    case default
      call error_stop(" Error in quark param file (algorithm)")
    end select
  enddo
  return
end subroutine

end module
