#include <config.h>

module main_hadron_class
  use config_profile_class
  use timer_class
  use comlib
  use filename
  use lattice_class
  use solver_parameter_class
  use inverse_class
  use logging_class
  use gauge_field_class
  use gauge_field_dd_class
  use single_prec_class
  use wilson_quark_field_class
  use wilson_quark_field_dd_class
  use wilson_quark_dd_class
#ifdef _CLOVER
  use s_cchlf_class
  use clover_quark_class
  use clover_quark_dd_class
#endif
  use wilson_quark_parameter_class
  use gauge_dd_class
  use quark_prop_class
  use wavefunc_class
  implicit none
  private

  public :: main_hadron_obj
  public :: new,read,print,delete

  type main_hadron_obj
    type(gvfield_dd_wg_obj),         pointer :: u => NULL()
    type(quark_prop_parameter_obj),  pointer :: quark_parameter(:) => NULL()
    type(wavefunc_param_obj),        pointer :: wavefunc(:) => NULL()
    type(config_profile_obj) :: config_prof
    character(len=_CHARLEN_) :: config_name,config_path,fname,input_fname,cwd,job_id
    character(len=_CHARLEN_) :: cdate,ctime,czone,comment,str
    character(len=_CHARLEN_) :: hadron_path
    integer :: num_quark, num_wave
    integer :: ini_traj,end_traj,traj_skip,traj
    type(logging_obj), pointer :: solver_log
    type(timer_obj) :: run_time
  end type

  interface new
    module procedure new_main
  end interface

  interface delete
    module procedure delete_main
  end interface

  interface read
    module procedure read_main
    module procedure read_config
  end interface

  interface print
    module procedure print_main
  end interface

contains

subroutine new_main(this)
  implicit none
  type(main_hadron_obj), intent(inout) :: this
  type(lattice_obj)   :: lattice
  type(gfields_obj)   :: gauge_fields
  type(wqfields_obj)  :: wilson_quark_fields
  type(gfields_dd_obj)  :: gauge_fields_dd
  type(wqfields_dd_obj) :: wilson_quark_fields_dd
#ifdef _CLOVER
  type(clfields_obj) :: clover_fields
  type(clfields_dd_obj) :: clover_fields_dd
#endif
  type(single_prec_obj) :: single_prec
  character(len=8)  :: cdate
  character(len=10) :: ctime
  character(len=_CHARLEN_) :: fname

  call new(lattice)

  call new(this%run_time)
  call tic(this%run_time)

  this%input_fname=repeat(' ',len(this%input_fname))
  this%cwd        =repeat(' ',len(this%cwd))
  this%job_id     =repeat(' ',len(this%job_id))

  call DATE_AND_TIME(date=cdate,time=ctime)
  write(this%job_id,'(A,1X,A)')TRIM(ADJUSTL(cdate)),TRIM(ADJUSTL(ctime))

  if (nodeid == 0) then
    call getarg(1,this%input_fname)
    call getcwd(this%cwd)
  endif

  if (nodeid == 0) then
    write(*,'(" DATE AND TIME :",A)')TRIM(this%job_id)
  endif

  if (NPU > 1) then
    call comlib_bcast(this%input_fname,0)
    call comlib_bcast(this%cwd,0)
    call comlib_bcast(this%job_id,0)
  endif

#ifdef _ALIGNCHECK_
!**********************************
!*  eo/deo fields 
!* (used in e/o-site prec UVPHMC)
!**********************************
  call new(gauge_fields)
  call new(wilson_quark_fields)
#ifdef _CLOVER
  call new(clover_fields)
#endif
#endif

!*******************
!*  deo/eo fields
!*******************
  call new(gauge_fields_dd)
  call new(wilson_quark_fields_dd)
#ifdef _CLOVER
  call new(clover_fields_dd)
#endif

!*****************************************
!*  deo/eo fields (single precision)
!*****************************************
  call new(single_prec)

  allocate(this%u)
  call new(this%u)

  allocate(this%solver_log)
  call new(this%solver_log,'quark_solver.log')

  return
end subroutine

subroutine delete_main(this)
  implicit none
  type(main_hadron_obj), intent(inout) :: this
  type(lattice_obj)   :: lattice
  integer :: iqk

  do iqk=1,this%num_wave
    call delete(this%wavefunc(iqk))
  enddo

  do iqk=1,this%num_quark
    call delete(this%quark_parameter(iqk))
  enddo
  call delete(this%solver_log)

  deallocate(this%wavefunc)
  deallocate(this%quark_parameter)
  deallocate(this%solver_log)

  call tac(this%run_time)
  call average(this%run_time)
  if (nodeid==0) then
    write(*,'(80("="))')
    write(*,'(" Hadron ETime:",E24.16)')this%run_time%elapse
  endif
  call delete(lattice)

  return
end subroutine

subroutine read_main(iout,this)
  implicit none
  integer, intent(in) :: iout
  type(main_hadron_obj), intent(inout) :: this
  integer :: iqk,iwf

  if (nodeid == 0) then
    open(iout,file=this%input_fname,form='formatted',action='read',status='old')
    read(iout,*)
    read(iout,*)this%ini_traj,this%end_traj,this%traj_skip
    read(iout,'(A)')this%config_path
    read(iout,'(A)')this%config_name
    read(iout,*)
    read(iout,'(A)')this%hadron_path
    read(iout,*)
    read(iout,*)this%num_wave
  endif

  if (NPU > 1) then
    call comlib_bcast(this%ini_traj,0)
    call comlib_bcast(this%end_traj,0)
    call comlib_bcast(this%traj_skip,0)
    call comlib_bcast(this%config_path,0)
    call comlib_bcast(this%config_name,0)
    call comlib_bcast(this%hadron_path,0)
    call comlib_bcast(this%num_wave,0)
  endif

  allocate(this%wavefunc(this%num_wave))
  do iqk=1,this%num_wave
    call new(this%wavefunc(iqk))
  enddo

  do iwf=1,this%num_wave
    if (nodeid == 0) then
      read(iout,*)
    endif
    call read(iout,this%wavefunc(iwf))
    call get(this%wavefunc(iwf))
  enddo

  if (nodeid == 0) then
    read(iout,*)
    read(iout,*)this%num_quark
  endif

  if (NPU > 1) then
    call comlib_bcast(this%num_quark,0)
  endif

  allocate(this%quark_parameter(this%num_quark))
  do iqk=1,this%num_quark
    this%quark_parameter(iqk)%solver_log => this%solver_log
    call new(this%quark_parameter(iqk))
  enddo

  do iqk=1,this%num_quark
    if (nodeid == 0) then
      read(iout,*)
    endif
    call read(iout,this%quark_parameter(iqk))
  enddo

  return
end subroutine

#include "read_config.h90"

subroutine print_main(this)
  implicit none
  type(main_hadron_obj), intent(inout) :: this
  type(lattice_obj) :: lattice
  integer :: iqk

  call print(lattice)

  if (nodeid==0) then
    write(*,'(80("="))')
    write(*,'("   Hadron Save path :"A)')TRIM(this%hadron_path)
    write(*,'(80("="))')
  endif
  
  return
end subroutine

end module
